/* $Id: cartanal.cpp,v 1.1.1.1 1996/04/22 14:03:20 cvs Exp $
 *
 * Copyright (c) 1996 University of Alberta
 *
 * May be copied or modified under the terms of the GNU General Public
 * License.  See COPYING file for more information.
 */

#include <owl\owlcore.h>
#include <owl\static.h>
#include <owl\opensave.h>
#include <owl\radiobut.h>
#include <owl\listbox.h>
#include <owl\edit.h>
#include <owl\validate.h>
#include <math.h>
#pragma hdrstop

#include <image.h>
#include "cartanal.h"
#include "cartanal.rh"

#define MAX(i, j) ((i)>(j) ? (i) : (j))
#define MIN(i, j) ((i)<(j) ? (i) : (j))
#define SQR(i) ((i)*(i))

DEFINE_RESPONSE_TABLE1 (TApp, TApplication)
  EV_COMMAND (CM_HELPINDEX, CmIndex),
//  EV_COMMAND (CM_HELPKEYBOARD, CmKeyboard),
//  EV_COMMAND (CM_HELPPROCEDURES, CmProcedures),
  EV_COMMAND (CM_HELPUSING_HELP, CmUsingHelp),
  EV_COMMAND (CM_ABOUT, CmAbout),
END_RESPONSE_TABLE;

DEFINE_RESPONSE_TABLE1 (TMainWindow, TFrameWindow)
  EV_COMMAND (CM_OPEN, CMOpen),
  EV_COMMAND (CM_ADDFILE, CMAddFile),
  EV_COMMAND (CM_REMOVEFILE, CMRemoveFile),
  EV_COMMAND (CM_PREFERENCES, CMPreferences),
  EV_COMMAND (CM_QUIT, CMQuit),
  EV_COMMAND (CM_STARTBATCH, CMStartBatch),
  EV_COMMAND (CM_STOPBATCH, CMStopBatch),
  EV_COMMAND (CM_SAVEENH, CMSaveEnhanced),
  EV_COMMAND (CM_ENHANCE, CMEnhanceImage),
  EV_COMMAND (CM_COUNTCELLS, CMCountCells),
  EV_COMMAND (CM_SPECIALHANDCOUNT, CMHandCount),
  EV_COMMAND (CM_COMPRESS, CMCompressImage),
  EV_WM_SIZE,
  EV_WM_TIMER,
END_RESPONSE_TABLE;

DEFINE_RESPONSE_TABLE1(TPreferencesDialog, TDialog)
  EV_COMMAND (IDOK, CmOk),
  EV_COMMAND (CM_DEFAULTS, CmDefaults),
END_RESPONSE_TABLE;

DEFINE_RESPONSE_TABLE1(THandCountDialog, TDialog)
  EV_COMMAND (IDOK, CmOk),
  EV_COMMAND (CM_UNDO, CmUndo),
  EV_CHILD_NOTIFY(IDC_SIZESMALL, BN_CLICKED, BnCircleSize),
  EV_CHILD_NOTIFY(IDC_SIZEMED, BN_CLICKED, BnCircleSize),
  EV_CHILD_NOTIFY(IDC_SIZELARGE, BN_CLICKED, BnCircleSize),
END_RESPONSE_TABLE;

DEFINE_RESPONSE_TABLE1( TDisplayWindow, TFrameWindow )
  EV_WM_RBUTTONDOWN,
  EV_WM_LBUTTONDOWN,
  EV_WM_MOUSEMOVE,
  EV_WM_LBUTTONUP,
END_RESPONSE_TABLE;


// Global Variables

int LastPercent;


//  Global Methods

void far StatusFunc (int PercentDone)
  {
  LastPercent = PercentDone;
  MSG Msg;
  if (PeekMessage (&Msg, NULL, 0, 0, PM_REMOVE))
	{
	TranslateMessage (&Msg);
	DispatchMessage (&Msg);
	}
  }

void InterlaceReduction (MemoryImage *image)
  {
  int x, y, w, h;
  int r1, g1, b, r2, g2;

  image->GetImageSize (w, h);
  for (y=0; y<h-1; y++)
	{
	StatusFunc ((long)100*y/h);
	for (x=0; x<w; x++)
	  {
	  image->GetPixelTC (x, y, r1, g1, b);
	  image->GetPixelTC (x, y+1, r2, g2, b);
	  image->PutPixelTC (x, y, (r1+r2)>>1, (g1+g2)>>1, b);
	  }
	}
  }

void HSICalculateIntensityHistogram(MemoryImage *image, long *intensity)
  {
  if (!intensity) return;
  int x, y, w, h, i;
  int r, g, b;

  for (i=0; i<256; i++) intensity[i]=0;
  image->GetImageSize (w, h);
  for (y=0; y<h-1; y++)
	{
	StatusFunc ((long)100*y/h);
	for (x=0; x<w; x++)
	  {
	  image->GetPixelTC (x, y, r, g, b);
//	  intensity[(r+g)/2]++;
	  intensity[MAX(r,g)]++;
	  }
	}
  }

void HSICalculateLocalIntensityHistogram(MemoryImage *image, long *intensities, TRect &rect)
  {
  if (!intensities) return;
  int x, y, w, h, i;
  int r, g, b;

  for (i=0; i<256; i++) intensities[i]=0;
  image->GetImageSize (w, h);
  for (y=rect.top; y<rect.bottom; y++)
	{
	StatusFunc ((long)100*y/h);
	for (x=rect.left; x<rect.right; x++)
	  {
	  image->GetPixelTC (x, y, r, g, b);
	  intensities[MAX(r,g)]++;
	  }
	}
  }

void HSICalculateSaturationHistogram(MemoryImage *image, long *saturation, int minInt=0)
  {
  if (!saturation) return;
  int x, y, w, h, i;
  int r, g, b, s;

  for (i=0; i<256; i++) saturation[i]=0;
  image->GetImageSize (w, h);
  for (y=0; y<h-1; y++)
	{
	StatusFunc ((long)100*y/h);
	for (x=0; x<w; x++)
	  {
	  image->GetPixelTC (x, y, r, g, b);
	  if (MAX(r,g) > minInt)
		{
		s = r-g;
		saturation[(s<0 ? -s : s)]++;
		}
	  }
	}
  }

void SmoothHistogram(long *hist)
  {
  int i;  long prev = hist[0], templ;
  for (i=1; i<255; i++)
	{
	templ = (prev + 2*hist[i] + hist[i+1]) / 4;
	prev = hist[i];
	hist[i] = templ;
	}
  hist[0] = (hist[1]+hist[0]) / 2;
  hist[255] = (hist[254]+hist[255]) / 2;
  }

void HSIEnhanceImage(MemoryImage *image, int zeroInt=0, int midInt=128, int maxInt=255, int satPercent=100)
  {
  if (!image) return;
  int x, y, w, h;
  int r, g, b;
  int i, rs, gs;

  image->GetImageSize (w, h);

  for (y=0; y<h-1; y++)
	{
	StatusFunc ((long)100*y/h);
	for (x=0; x<w; x++)
	  {
	  image->GetPixelTC(x, y, r, g, b);

//	  i = (r+g)/2;
	  i = MAX(r,g);
	  rs = r-i;
	  gs = g-i;

	  if (i<midInt) i = (i-zeroInt)*128/(midInt-zeroInt);
		else i = (i-midInt)*128/(maxInt-midInt) + 128;

	  r = i + rs*satPercent/100;
	  g = i + gs*satPercent/100;

	  if (r<0) r=0;  if (r>255) r=255;
	  if (g<0) g=0;  if (g>255) g=255;
//	  if (b<0) b=0;  if (b>255) b=255;

	  image->PutPixelTC(x, y, r, g, b);
	  }
	}
  }

void HSIEnhanceImageLocal(MemoryImage *image, int *zeroInt, int *midInt, int maxInt, int satPercent, int numCol, int numRow)
  {
  if (!image) return;
  int x, y, w, h;
  int r, g, b;
  int i, rs, gs;
  int j=0, k;
  int x1, x2, y1, y2;
  int zi=0, mi=128;

  image->GetImageSize(w, h);

  for (y=0; y<h; y++)
	{
	StatusFunc ((long)100*y/h);

	y1=(2*j-1)*h/numRow/2;  y2=(2*j+1)*h/numRow/2;
	if (y>y2)
	  {  j++;  y1=y2;  y2=(2*j+1)*h/numRow/2;  }

	k=0;
	x1=(2*k-1)*w/numCol/2;  x2=(2*k+1)*w/numCol/2;
	for (x=0; x<w; x++)
	  {
	  if (x>x2)
		{  k++;  x1=x2;  x2=(2*k+1)*w/numCol/2;  }

	  int sumZero=0, sumMid=0, n=0, d;
	  if ((k>0)&&(j>0)&&(x!=x2)&&(y!=y2))
		{ d=0x100000/(SQR(x1-x)+SQR(y1-y)+1); n+=d;
		  sumZero+=d*zeroInt[(j-1)*numCol+k-1];
		  sumMid+=d*midInt[(j-1)*numCol+k-1]; }
	  if ((k>0)&&(j<numRow)&&(x!=x2))
		{ d=0x100000/(SQR(x1-x)+SQR(y2-y)+1); n+=d;
		  sumZero+=d*zeroInt[j*numCol+k-1];
		  sumMid+=d*midInt[j*numCol+k-1]; }
	  if ((k<numCol)&&(j>0)&&(y!=y2))
		{ d=0x100000/(SQR(x2-x)+SQR(y1-y)+1); n+=d;
		  sumZero+=d*zeroInt[(j-1)*numCol+k];
		  sumMid+=d*midInt[(j-1)*numCol+k]; }
	  if ((k<numCol)&&(j<numRow))
		{ d=0x100000/(SQR(x2-x)+SQR(y2-y)+1); n+=d;
		  sumZero+=d*zeroInt[j*numCol+k];
		  sumMid+=d*midInt[j*numCol+k]; }

	  if (n>0) zi=sumZero/n, mi=sumMid/n;
	  if (mi==maxInt) mi--;
	  if (zi>=mi) zi=mi-1;

	  image->GetPixelTC(x, y, r, g, b);

	  i = MAX(r,g);
	  rs = r-i;
	  gs = g-i;

	  if (i<mi) i = (i-zi)*128/(mi-zi);
		else i = (i-mi)*128/(maxInt-mi) + 128;

	  r = i + rs*satPercent/100;
	  g = i + gs*satPercent/100;

	  if (r<0) r=0;  if (r>255) r=255;
	  if (g<0) g=0;  if (g>255) g=255;
//	  if (b<0) b=0;  if (b>255) b=255;

	  image->PutPixelTC(x, y, r, g, b);
	  }
	}
  }

int maxValue (MemoryImage *image, int x, int y)
  {
  int r, g, b;
  image->GetPixelTC (x, y, r, g, b);
  return (r>g ? r : g);
  }

// obsolete functions
void RedReduction (MemoryImage *image)
  {
  int x, y, r, g, b, w, h;
  long l;

  image->GetImageSize (w, h);
  for (y=0; y<h; y++)
	{
	StatusFunc ((long)100*y/h);
	for (x=0; x<w; x++)
	  {
	  image->GetPixelTC (x, y, r, g, b);
	  l = g-128;
	  if (l>0) r = r - ((l * r) >> 9);
	  if (r<0) r = 0;
	  image->PutPixelTC (x, y, r, g, b);
	  }
	}
  }


//  TMainWindow methods

TMainWindow::TMainWindow (TWindow *AParent = 0) :
	 TFrameWindow (AParent, "Cartilage Image Analysis")
  {
  Attr.W = 350;  Attr.H = 250;
  AssignMenu (id_MAINMENU);
  FuncText = new TStatic (this, 98, "", 10, 5, 230, 24, 40);
  StatText = new TStatic (this, 99, "", 250, 5, 150, 24, 40);
  StackText = new TStatic (this, 97, "", 10, 29, 300, 20, 40);
  LastPercent=numCells=-1;  

  fileNameLB = new TListBox (this, id_FILENAMELB, 0, LByofs, 50, 50);
  fileNameLB->Attr.Style |= LBS_NOINTEGRALHEIGHT;
  FileData = new TOpenSaveDialog::TData(OFN_HIDEREADONLY|OFN_FILEMUSTEXIST,
	 "All Images|*.tga;*.bmp;*.dib;*.enh|Targa Images (*.TGA)|*.tga|Bitmap Images (*.BMP;*.DIB)|*.bmp;*.dib|Enhanced Images (*.ENH)|*.enh|All Files (*.*)|*.*",
	 0, "", "BMP");
  *FileData->FileName = 0;

  prefs.LoadPreferences();
  const char *iniFile = prefs.GetLocalIniFileName();
//  FileData->FilterIndex = GetPrivateProfileInt(CartanalSection, "FilterIndex", 1, iniFile);
  Attr.X = GetPrivateProfileInt(CartanalSection, "WindowPosX", Attr.X, iniFile);
  Attr.Y = GetPrivateProfileInt(CartanalSection, "WindowPosY", Attr.Y, iniFile);

  busy = wantToQuit = wantToStop = doingBatch = 0;
  saveEnhanced = prefs.defSaveEnhanced;
  DisplayWindow = NULL;
  image = NULL;  imageType = 0;
  strcpy (currentFile, "");
  }

TMainWindow::~TMainWindow ()
  {
  KillTimer(101);
  prefs.SavePreferences();

  const char *iniFile = prefs.GetLocalIniFileName();
  char s[16];
//  wsprintf(s, "%d", FileData->FilterIndex);
//  WritePrivateProfileString(CartanalSection, "FilterIndex", s, iniFile);
  wsprintf(s, "%d", Attr.X);
  WritePrivateProfileString(CartanalSection, "WindowPosX", s, iniFile);
  wsprintf(s, "%d", Attr.Y);
  WritePrivateProfileString(CartanalSection, "WindowPosY", s, iniFile);

  delete FileData;

  if (image) delete image;
  WinHelp(HLPFileName, HELP_QUIT, 0);
  }

int TMainWindow::CanClose ()
  {
  if (busy | doingBatch)
	{
	wantToQuit = 1;
	EnableMenuItem (GetMenu(), CM_STOPBATCH, MF_BYCOMMAND | MF_GRAYED);
	return 0;
	}
  if (DisplayWindow)
	{
	DisplayWindow->Destroy();
	delete DisplayWindow;
	DisplayWindow = NULL;
	}
  return 1;
  }

void TMainWindow::SetupWindow ()
  {
  TFrameWindow::SetupWindow();

  SetTimer(101, StatusInterval);
  EnableMenuItem (GetMenu(), CM_STOPBATCH, MF_BYCOMMAND | MF_GRAYED);
  if (saveEnhanced)
	CheckMenuItem(GetMenu(), CM_SAVEENH, MF_BYCOMMAND | MF_CHECKED);
	else CheckMenuItem(GetMenu(), CM_SAVEENH, MF_BYCOMMAND | MF_UNCHECKED);

  DrawMenuBar();
  SetClassLong (GCL_HICON, (unsigned short)GetModule()->LoadIcon(icn_MAIN));
  }

void TMainWindow::CMOpen()
  {
  if (busy) return;
  busy++;
  *FileData->FileName = 0;
  if (TFileOpenDialog (this, *FileData).Execute() == IDOK)
	{
	strlwr (FileData->FileName);
	if (strcmp(FileData->FileName, currentFile) || (imageType != 1))
	  {
	  strcpy (currentFile, FileData->FileName);  imageType = 0;

	  if (OpenImageFile(currentFile))
		SetStatus ("Image Loaded");
	  }
	}
  busy--;
  }

int TMainWindow::OpenImageFile(char *filename)
  {
  if (image) delete image;
  image = new MemoryImage (8, 8, 8);
  if (!image)
	{ SetStatus("Error Creating Image (101)");
	  return 0;  }
  image->SetStatusFunction (StatusFunc);

  SetStatus ("Loading Image", 0);
  ImageFile *inimage = DetermineImageTypeAndOpen(filename);

  if ((!inimage) || (inimage->bad()))
	{ SetStatus("Error Reading File (102)");
	  delete inimage; return 0;  }
  if (!image->LoadImage (inimage))
	{ SetStatus("Insufficient Memory (103)");  return 0;  }
  imageType = 1;

  BMPFile *bmp = TYPESAFE_DOWNCAST(inimage, BMPFile);
  TGAFile *tga = TYPESAFE_DOWNCAST(inimage, TGAFile);
  if (((bmp)&&(bmp->GetSpecialFileType()==specialEnhancedType)) ||
	  ((tga)&&(strstr(filename, EnhExtension))))
	{
	imageType = 2;
//	strcpy (strrchr (currentFile, '.'), TGAExtension);
	}

  delete inimage;
  SetStatus ("Displaying", 0);
  DisplayImage();
  return 1;
  }

void TMainWindow::CMAddFile()
  {
  if (busy) return;
  char path[fsPathName];
  char far *fname, *nextname;
  *FileData->FileName = 0;
  FileData->Flags |= OFN_ALLOWMULTISELECT;
  while (TFileOpenDialog (this, *FileData, 0, "Add File").Execute() == IDOK)
	{
	strlwr (FileData->FileName);
	fname = strchr(FileData->FileName, 32);
	if (fname) {  *fname = 0;  fname++;  }
	if ((!fname) && (fileNameLB->FindExactString (FileData->FileName, -1) < 0))
	  fileNameLB->AddString(FileData->FileName);
	while (fname)
	  {
	  strcpy(path, FileData->FileName);
	  strcat(path, "\\");
	  nextname = strchr(fname, 32);
	  if (nextname) {  *nextname = 0;  nextname++;  }
	  strcat(path, fname);
	  fname = nextname;

	  if (fileNameLB->FindExactString (path, -1) < 0)
	fileNameLB->AddString (path);
	  }
	strcpy (FileData->FileName, "");
	}
  FileData->Flags &= ~OFN_ALLOWMULTISELECT;
  }

void TMainWindow::CMRemoveFile()
  {
  if (fileNameLB->GetSelIndex() >= 0)
	fileNameLB->DeleteString (fileNameLB->GetSelIndex());
	else MessageBox ("No File Selected", "ERROR", MB_OK | MB_ICONEXCLAMATION);
  }

void TMainWindow::CMPreferences()
  {
  TPreferencesDialog(this, &prefs).Execute();
  }

void TMainWindow::CMQuit()
  {  CloseWindow ();  }

void TMainWindow::CMStartBatch()
  {
  if ((busy) || (doingBatch)) return;
  doingBatch++;
  wantToQuit = wantToStop = 0;
  EnableMenuItem (GetMenu(), CM_STARTBATCH, MF_BYCOMMAND | MF_GRAYED);
  EnableMenuItem (GetMenu(), CM_STOPBATCH, MF_BYCOMMAND | MF_ENABLED);
  DrawMenuBar();
  if ((image) && (imageType == 2)) CMCountCells();

  while ((fileNameLB->GetCount() > 0) && (!wantToQuit) && (!wantToStop))
	{
	if (imageType != 1)
	  {
	  if (image) delete image;
	  image = NULL;
	  imageType = 0;
	  fileNameLB->GetString (currentFile, 0);  fileNameLB->DeleteString (0);
	  }
	CMEnhanceImage();
	if ((!wantToQuit) && (!wantToStop)) CMCountCells();
	};
  if ((wantToQuit) || (wantToStop)) SetStatus ("Batch Stopped");
	else SetStatus ("Batch Complete");
  EnableMenuItem (GetMenu(), CM_STARTBATCH, MF_BYCOMMAND | MF_ENABLED);
  EnableMenuItem (GetMenu(), CM_STOPBATCH, MF_BYCOMMAND | MF_GRAYED);
  DrawMenuBar();
  doingBatch--;
  if (wantToQuit) CloseWindow();
  }

void TMainWindow::CMStopBatch()
  {
  if (doingBatch) wantToStop = 1;
  EnableMenuItem (GetMenu(), CM_STOPBATCH, MF_BYCOMMAND | MF_GRAYED);
  DrawMenuBar();
  }

void TMainWindow::CMSaveEnhanced()
  {
  (saveEnhanced ? saveEnhanced=0 : saveEnhanced=1);
  if (saveEnhanced)
	CheckMenuItem(GetMenu(), CM_SAVEENH, MF_BYCOMMAND | MF_CHECKED);
	else CheckMenuItem(GetMenu(), CM_SAVEENH, MF_BYCOMMAND | MF_UNCHECKED);
  }

void TMainWindow::CMEnhanceImage()
  {
  if (busy) return;
  StackText->Clear();
  if (prefs.useOldEnhance) {CMOldEnhanceImage(); return;}
  if (!strlen(currentFile)) CMOpen();
  if (!strlen(currentFile)) return;
  busy++;
  if ((!image) || (imageType != 1))
	{
	if (!OpenImageFile(currentFile))
	  { busy--; return; }
	if (imageType!=1)
	  { SetStatus("Image Already Enhanced (201)");
		busy--; return; }
	}

  SetStatus ("Interlace Reduction", 0);
  InterlaceReduction(image);
  SetStatus ("Smoothing Red", 0);
  image->Smooth (red, 3);
  if (prefs.extraSmooth) image->Smooth (red, 5);
  SetStatus ("Smoothing Green", 0);
  image->Smooth (green, 3);
  if (prefs.extraSmooth) image->Smooth (green, 5);

  SetStatus ("Displaying", 0);
  DisplayImage ();

  SetStatus ("Calculating Intensity", 0);
  long intensities[256];
  int intThres[64], intLow[64];

  HSICalculateIntensityHistogram(image, intensities);
  SmoothHistogram(intensities);
  SmoothHistogram(intensities);
  int i, intMax=0;
  for (i=1; i<255; i++) if (intensities[i]>0) intMax=i;

  int width, height;
  image->GetImageSize(width, height);
  int numRow=1, numCol=1;
  if (prefs.localization==2) numRow=height/100, numCol=width/100;
  else if (prefs.localization==1) numRow=height/200, numCol=width/200;
  if (numRow==0) numRow=1;  if (numRow>8) numRow=8;
  if (numCol==0) numCol=1;  if (numCol>8) numCol=8;

  int j,k;
  for(j=0; j<numRow; j++)
	for(k=0; k<numCol; k++)
	  {
	  TRect rect;
	  rect.top = height*j/numRow;	rect.bottom = height*(j+1)/numRow;
	  rect.left = width*k/numCol; 	rect.right = width*(k+1)/numCol;
	  HSICalculateLocalIntensityHistogram(image, intensities, rect);
	  SmoothHistogram(intensities);
	  SmoothHistogram(intensities);

	  int intMode;
	  long maxFreq=0;
	  for (i=1; i<255; i++)
		if (intensities[i]>maxFreq) intMode=i, maxFreq=intensities[i];

	  int bot=intMode-1, top=intMode+1;
	  unsigned long n=intensities[intMode], varsum=0;

	  do {
		long sum=0, sumcount=0;
		if (bot>0) sum+=intensities[bot], sumcount++;
		if (top<255) sum+=intensities[top], sumcount++;
		if (sumcount>0)
		  {
		  varsum += 2*sum/sumcount*(top-intMode)*(top-intMode)/100;
		  n += 2*sum/sumcount;
		  }
		bot--; top++;
		} while ((bot>0)||(top<255));

	  int stdev = varsum>0 ? sqrt(varsum/(n/100)) : 0;
	  intThres[j*numCol+k] = intMode + stdev*prefs.backgroundStDevs/10;

	  // why is this line here?  if thres > max intensity, set thres = avg(max,mode)
	  if (intThres[j*numCol+k]>=intMax) intThres[j*numCol+k]=(intMax+intMode)/2;

	  intLow[j*numCol+k] = intMode;
	  }

  int thres=0, n=0;
  for(j=0; j<numRow; j++) for(k=0; k<numCol; k++)
	thres+=intThres[j*numCol+k], n++;

  if (prefs.displayHistogram)
	{
	TWindow *f = new THistogramWindow(this, "Intensity Histogram", intensities, 0, thres/n);
	f->Create();
	}

  SetStatus ("Calculating Saturation", 0);
  long saturation[256];
  HSICalculateSaturationHistogram(image, saturation, thres/n);
  SmoothHistogram(saturation);

  long templ=0;  n=0;
  for (i=0; i<256; i++)
	n += saturation[i], templ += i*saturation[i];
  int satMean = (n ? templ*3/n : 128);
  if (prefs.displayHistogram)
	{
	TWindow *f = new THistogramWindow(this, "Saturation Histogram", saturation, satMean);
	f->Create();
	}

  SetStatus ("Enhancing Image", 0);
  HSIEnhanceImageLocal(image, intLow, intThres, intMax, 100*prefs.idealSat/satMean, numCol, numRow);

  SetStatus ("Smoothing Red", 0);
  image->Smooth (red, 3);
  image->Smooth (red, prefs.extraSmooth ? 5 : 3);
  SetStatus ("Smoothing Green", 0);
  image->Smooth (green, 3);
  image->Smooth (green, prefs.extraSmooth ? 5 : 3);

  SetStatus ("Displaying", 0);
  DisplayImage ();

  imageType = 2;
  if ((!doingBatch)||(saveEnhanced))
	{
	SetStatus ("Saving Image", 0);
	HCURSOR oldCursor = ::SetCursor (LoadCursor (NULL, IDC_WAIT));
	char ENHname[fsPathName];
	strcpy (ENHname, currentFile);
	strcpy (strrchr (ENHname, '.'), EnhExtension);

	BMPFile *outimage = new BMPFile (ENHname, IMAGE_WRITE);

	if ((!outimage) || (outimage->bad()))
	  { SetStatus("Error Writing File (202)");
		delete outimage;
		busy--;  return;  }
//	outimage->SetPixelSize (24, 1);
	outimage->SetPixelSize (16, 1);
	outimage->SetCompression(BMP_COMPRESSION_BITFIELDS, 8, 8, 0);
	outimage->SetSpecialFileType(specialEnhancedType);
	outimage->LoadImage (image);
	delete outimage;
	::SetCursor (oldCursor);
	}

  SetStatus ("Enhancement Complete");
  busy--;
  if ((wantToQuit) && (!doingBatch)) CloseWindow();
  }

void TMainWindow::CMOldEnhanceImage()
  {
  if (busy) return;
  if (!strlen(currentFile)) CMOpen();
  if (!strlen(currentFile)) return;
  busy++;
  if ((!image) || (imageType != 1))
	{
	if (!OpenImageFile(currentFile))
	  { busy--; return; }
	if (imageType!=1)
	  { SetStatus("Image Already Enhanced (211)");
		busy--; return; }
	}

  SetStatus ("Smoothing", 0);
  image->Smooth (green, 3);
  image->Smooth (red, 3);
  SetStatus ("Red Reduction", 0);
  RedReduction (image);
  SetStatus ("Smoothing", 0);
  image->Smooth (green, 5);
  image->Smooth (red, 5);

  SetStatus ("Displaying", 0);
  DisplayImage ();

  SetStatus ("Calculating Histogram", 0);
  image->CalculateHistogram();
  image->SmoothHistogram();

  SetStatus ("Contrast Enhancement", 0);
  int mr = image->GetMaximumIntensity(red);
  int mg = image->GetMaximumIntensity(green);
  image->EnhanceContrast (red, image->GetBackgroundIntensity (red), MIN(mg,mr));
  image->EnhanceContrast (green, image->GetBackgroundIntensity (green), MAX(mr,mg));

  SetStatus ("Smoothing", 0);
  image->Smooth (green, 3);
  image->Smooth (red, 3);
  SetStatus ("Shifting Intensities", 0);
  image->ShiftIntensity (red, 1);
  SetStatus ("Displaying", 0);
  DisplayImage ();

  imageType = 2;
  if ((!doingBatch) || (saveEnhanced))
	{
	SetStatus ("Saving Image", 0);
	HCURSOR oldCursor = ::SetCursor (LoadCursor (NULL, IDC_WAIT));
	char ENHname[fsPathName];
	strcpy (ENHname, currentFile);
	strcpy (strrchr (ENHname, '.'), EnhExtension);
	BMPFile *outimage = new BMPFile (ENHname, ImageFile::WRITE);
	if ((!outimage) || (outimage->bad()))
	  { SetStatus("Error Writing File (212)");
		delete outimage;
		busy--;  return;  }
	outimage->SetPixelSize (24, 1);
	outimage->SetSpecialFileType(specialEnhancedType);
	outimage->LoadImage (image);
	delete outimage;
	::SetCursor (oldCursor);
	}
  SetStatus ("Enhancement Complete");
  busy--;
  if ((wantToQuit) && (!doingBatch)) CloseWindow();
  }

void TMainWindow::CMCompressImage()
  {
  if (busy) return;
  StackText->Clear();
  if (!strlen(currentFile)) CMOpen();
  if (!strlen(currentFile)) return;
  busy++;

  if ((!image)||(imageType!=1))
	{
	if (!OpenImageFile(currentFile))
	  { busy--; return; }
	if (imageType!=1)
	  { SetStatus("Image Already Compressed (401)");
		busy--; return; }
	}

  SetStatus ("Saving Image", 0);
  HCURSOR oldCursor = ::SetCursor (LoadCursor (NULL, IDC_WAIT));
  BMPFile *outimage = new BMPFile (currentFile, ImageFile::WRITE);
  if ((!outimage) || (outimage->bad()))
	{ SetStatus("Error Writing File (402)");
	  delete outimage;
	  busy--;  return;  }
  outimage->SetPixelSize(16, 1);
  outimage->SetCompression(BMP_COMPRESSION_BITFIELDS, 8, 8, 0);
  outimage->CopyFrom(image);
  delete outimage;
  ::SetCursor (oldCursor);
  SetStatus ("Compression Complete");
  busy--;
  if ((wantToQuit) && (!doingBatch)) CloseWindow();
  }

void TMainWindow::CMHandCount()
  {
  if (busy) return;
  StackText->Clear();
  if (!strlen(currentFile)) CMOpen();
  if (!strlen(currentFile)) return;
  busy++;

  if ((!image)||(imageType<1)||(imageType>2))
	{
	if (!OpenImageFile(currentFile))
	  { busy--; return; }

	SetStatus ("Displaying", 0);
	DisplayImage();
	}

  if (!DisplayWindow)
	{ SetStatus("Bad Display Window (311)");
	  busy--;  return;  }

  char DATname[fsPathName], datFile[fsPathName];
  strcpy (DATname, currentFile);
  if (strrchr(DATname, '\\')) strcpy(DATname, strrchr(DATname, '\\')+1);
  if (strrchr(DATname, ':'))  strcpy(DATname, strrchr(DATname, ':')+1);
  strcpy(strrchr (DATname, '.'), HandDataExtension);

  strcpy(datFile, prefs.DATPath);
  if ((strlen(datFile) != 0) && (datFile[strlen(datFile)-1] != '\\') &&
	  (datFile[strlen(datFile)-1] != ':'))
	strcat(datFile, "\\");
  strcat(datFile, DATname);

  fout = new ofstream (datFile);
  if ((!fout) || (fout->bad()))
	{ SetStatus("Error Writing File (312)");
	  delete fout;
	  busy--;  return;  }
  *fout << "hand counting...\n";
  *fout << "x" << "\t" << "y" << "\t" << "tRed" << "\t" << "tGreen" << "\t";
  *fout << "pixels" << "\t" << "red/green" << "\n";

  numCells=0;
  SetStatus ("Counting Cells");

  DisplayWindow->StartHandCount();
  }

void TMainWindow::DoneHandCounting(TCellPoints *cellPoints)
  {
  TCellPointsIterator j(*cellPoints);
  while (j)
	{
	TPoint p = j++;
	int x = (p.x<0 ? ~p.x : p.x);
	*fout << x << "\t" << p.y << "\t";
	if (p.x<0) *fout << "2\t1\t1\t2\n";
	else *fout << "1\t2\t1\t0.5\n";
	}

  numCells = -1;
  *fout << "Done." << "\n";
  delete fout;
  SetStatus ("Counting Complete");
  imageType = 3;
  busy--;
  if ((wantToQuit) && (!doingBatch)) CloseWindow();
  }

void TMainWindow::CMCountCells()
  {
  if (busy) return;
  StackText->Clear();
  int oldIndex = FileData->FilterIndex;
  FileData->FilterIndex = 4;
  if (!strlen(currentFile)) CMOpen();
  FileData->FilterIndex = oldIndex;
  if (!strlen(currentFile)) return;
  busy++;

  if ((!image)||(imageType!=2))
	{
	char ENHname[fsPathName];
	strcpy (ENHname, currentFile);
	strcpy (strrchr (ENHname, '.'), EnhExtension);

	if (!OpenImageFile(ENHname))
	  { busy--; return; }
	if (imageType!=2)
	  { SetStatus("Image Not Enhanced (301)");
		busy--; return; }

	SetStatus ("Displaying", 0);
	DisplayImage();
	}

  image->GetImageSize (W, H);

  char DATname[fsPathName], datFile[fsPathName];
  strcpy (DATname, currentFile);
  if (strrchr(DATname, '\\')) strcpy(DATname, strrchr(DATname, '\\')+1);
  if (strrchr(DATname, ':'))  strcpy(DATname, strrchr(DATname, ':')+1);
  strcpy(strrchr (DATname, '.'), DataExtension);

  strcpy(datFile, prefs.DATPath);
  if ((strlen(datFile) != 0) && (datFile[strlen(datFile)-1] != '\\') &&
	  (datFile[strlen(datFile)-1] != ':'))
	strcat(datFile, "\\");
  strcat(datFile, DATname);

  fout = new ofstream (datFile);
  if ((!fout) || (fout->bad()))
	{ SetStatus("Error Writing File (302)");
	  delete fout;
	  busy--;  return;  }
  *fout << "machine counting...";
  if (prefs.maxPixs!=0)
	*fout << " maxPixs=" << prefs.maxPixs;
  *fout << "\n";
  *fout << "x" << "\t" << "y" << "\t" << "tRed" << "\t" << "tGreen" << "\t";
  *fout << "pixels" << "\t" << "red/green" << "\n";

  numCells=0;
  SetStatus ("Counting Cells");
  int x, y, c, i;

  for (x=0; x<W; x++)
	{
	image->PutPixelTC (x, 0, 0, 0, 0);
	image->PutPixelTC (x, H-1, 0, 0, 0);
	}
  for (y=0; y<H; y++)
	{
	image->PutPixelTC (0, y, 0, 0, 0),
	image->PutPixelTC (W-1, y, 0, 0, 0);
	}

  x = 0;  y = 0;
  stackcount = stackmax = 0;
  do {
	FindMax (x, y, (x+blocksize < W ? x+blocksize : W-1),
		   (y+blocksize < H ? y+blocksize : H-1));
	x += blocksize;
	if (x>=W) x=0, y+=blocksize;
	} while (y<H);

  if (prefs.displayStackUsage)
	{
	wsprintf(stackstr, "max:%d", stackmax);
	StackText->SetText(stackstr);
	}

  numCells = -1;
  SetStatus ("Displaying", 0);
  DisplayImage ();
  *fout << "Done." << "\n";
  delete fout;
  SetStatus ("Counting Complete");
  imageType = 3;
  busy--;
  if ((wantToQuit) && (!doingBatch)) CloseWindow();
  }

void TMainWindow::DisplayImage ()
  {
  if (!DisplayWindow)
	{
	DisplayWindow = new TDisplayWindow (NULL);
	if (!DisplayWindow)
	  {  MessageBox("Unable To Create Display Window", "ERROR", MB_OK | MB_ICONEXCLAMATION);
	  return;  }
	DisplayWindow->Create();
	}
  DisplayWindow->DisplayImage(currentFile, image);

  char LINname[fsPathName];
  strcpy (LINname, currentFile);
  strcpy (strrchr (LINname, '.'), LineExtension);
  DisplayWindow->SetLineSaveName(LINname);
  }

void TMainWindow::FindMax(short int x1, short int y1, short int x2, short int y2)
  {
  stackcount++;  if (stackcount > stackmax) stackmax = stackcount;
  if (prefs.displayStackUsage)
	{
	wsprintf(stackstr, "stack:%d max:%d\tx:%d y:%d", stackcount, stackmax, x1, y1);
	StackText->SetText(stackstr);
	}

  static int r, g, b, c;
  short int mx, my, x, y;

  if (x2-x1 < blocksize) x1 = x2-blocksize;
  if (y2-y1 < blocksize) y1 = y2-blocksize;
  if (x1 < 0) x2 -= x1, x1 = 0;
  if (y1 < 0) y2 -= y1, y1 = 0;
  do {
	c = 0;
	for (y=y1; y<=y2; y++)
	  {
	  StatusFunc (-1);
	  for (x=x1; x<=x2; x++)
		{
		image->GetPixelTC (x, y, r, g, b);
		if (MAX(r,g) > c)
		  {
		  c = MAX(r,g);
		  mx = x;  my = y;
		  }
		}
	  }
	if (c > prefs.thres)
	  {
	  if ((mx > x1+edgesize) && (mx < x2-edgesize)
		  && (my > y1+edgesize) && (my < y2-edgesize))
		FindObject (mx, my);
	  else
		if (((mx<=edgesize) && (x1==0)) || ((my<=edgesize) && (y1==0)) ||
			((mx>=W-edgesize-1) && (x2==W-1)) || ((my>=H-edgesize-1) && (y2==H-1)))
		  FindObject (mx, my);
		else FindMax ((mx>nextBlockSize ? mx-nextBlockSize : 0),
				  (my>nextBlockSize ? my-nextBlockSize : 0),
				  (mx+nextBlockSize<W ? mx+nextBlockSize : W-1),
				  (my+nextBlockSize<H ? my+nextBlockSize : H-1));
	  } else c=0;
	} while (c > prefs.thres);

  stackcount--;
  if (prefs.displayStackUsage)
	{
	wsprintf(stackstr, "stack:%d max:%d\tx:%d y:%d", stackcount, stackmax, x1, y1);
	StackText->SetText(stackstr);
	}
  }

void TMainWindow::FindObject (int x, int y)
  {
  int seg;
  int radii[numSegments];
  for (seg=0; seg < numSegments; seg++)
	radii[seg] = FindEdge(x, y, 1000*cos(2*Pi*seg/numSegments), 1000*sin(2*Pi*seg/numSegments));

  int r0, r1, r2, i;
  for (i=0; i<numSmooths; i++)
	{
	r1=radii[0];
	r2 = r0 = radii[1];
	for (seg=1; seg < numSegments; seg++)
	  {
	  radii[seg] = (radii[(seg+1)%numSegments]+r1+r2)/3;
	  r1=r2;  r2=radii[(seg+1)%numSegments];
	  }
	radii[0] = (r0+r1+r2)/3;
	}

  int x1, x2, y1, y2;
  if (radii[0]==0) radii[0] = 1;
  x1 = x + radii[0];
  y1 = y;
  for (seg=1; seg < numSegments; seg++)
	{
	if (radii[seg]==0) radii[seg] = 1;
	x2 = x + (cos(2*Pi*seg/numSegments)*radii[seg]);
	y2 = y + (sin(2*Pi*seg/numSegments)*radii[seg]);
	Line (x1, y1, x2, y2);
	x1=x2;  y1=y2;
	}
  Line (x1, y1, x+radii[0], y);

  long tr=0, tg=0;
  radius=prefs.maxPixs?2+(prefs.maxPixs/3.1416):0;
  startx=x; starty=y; numcolor=0;

  long n=Fill (x, y, tr, tg);

  if (n >= prefs.minCellSize)
	{
	image -> PutPixelTC (x, y, prefs.thres-2, prefs.thres-2, 255);

	*fout << x << "\t" << y << "\t";
	*fout << tr << "\t" << tg << "\t" << n << "\t";

if ((prefs.maxPixs!=0)&&(numcolor<prefs.maxPixs))
  *fout << "*";

	if (tg > 0)
	  {
	  fout->setf (ios::fixed, ios::floatfield);
	  fout->precision (2);
	  *fout << (float)tr/tg << "\n";
	  }
	  else *fout << "all red" << "\n";

	DisplayWindow->Circle (x, y);
	char ch[40];  numCells++;
	}
  }

int TMainWindow::FindEdge (int x, int y, int cx, int cy)
  {
  int i=1, x1, y1;
  int c, c1, c2, mi=0, mc=1000;

  if ((x<2) && (cx < 0)) return 0;
  if ((y<2) && (cy < 0)) return 0;
  if ((x>W-3) && (cx > 0)) return 0;
  if ((y>H-3) && (cy > 0)) return 0;

  c1 = maxValue (image, x, y);
  x1 = x + ((long)cx*i)/1000;
  y1 = y + ((long)cy*i)/1000;
  c = maxValue (image, x1, y1);

  do {
	if (c==0)  mi=i,  mc=0;
	i++;
	c2 = c1; c1 = c;
	x1 = x + ((long)cx*i)/1000;
	y1 = y + ((long)cy*i)/1000;
	c = maxValue (image, x1, y1);
	if (c+c1+c2 < mc)  mc=c+c1+c2, mi=i-1;
	} while ((3*(prefs.thres-overShoot) < c+c1+c2) && (c+c1+c2 < mc+3*maxVar) &&
	(x1>0) && (y1>0) && (x1<W-1) && (y1<H-1));

  return mi;
  }

void TMainWindow::Line (int x1, int y1, int x2, int y2)
  {
  if ((x1==x2) && (y1==y2)) return;
  int d, x, y, ax, ay, sx=0, sy=0, dx, dy;

  dx = x2-x1;  ax = abs(dx) << 1;  if (dx) sx = abs(dx)/dx;
  dy = y2-y1;  ay = abs(dy) << 1;  if (dy) sy = abs(dy)/dy;
  x = x1;  y = y1;

  if (ax > ay)
	{
	d = ay - (ax >> 1);
	do {
	  image->PutPixelTC (x, y, 0, 0, 0);
	  if (x != x2)
	{
	if (d >= 0)  y = y + sy,  d = d - ax;
	x = x + sx,  d = d + ay;
	}
	  } while (x != x2);
	} else
    {
	d = ax - (ay >> 1);
	do {
	  image->PutPixelTC (x, y, 0, 0, 0);
	  if (y != y2)
	{
	if (d >= 0) x = x + sx,  d = d - ay;
	y = y + sy,  d = d + ax;
	}
	  } while (y != y2);
	}
  }

long TMainWindow::DoLine (int x, int y, long &tr, long &tg, long &sx, long &sy)
  {
  int x1=x, x2;
  int r, g, b;
  if (maxValue (image, x, y)==0) return 0;
  while ((maxValue (image, x1-1, y) > 0)) x1--;
  x=x1;
  while ((maxValue (image, x, y) > 0))
	{
	image->GetPixelTC (x, y, r, g, b);
	if ((radius==0)||((SQR(startx-x)+SQR(starty-y)<radius)&&(numcolor<prefs.maxPixs)))
	  tr+=r, tg+=g, sx+=(long)x*(r+g), sy += (long)y*(r+g), numcolor++;
	image->PutPixelTC (x, y, 0, 0, 0);
	x++;
	}
  x2=x;
  long n=0;
  for (x=x1; x<x2; x++)
	{
	if (maxValue (image, x, y+1) > 0)
	  n += DoLine (x, y+1, tr, tg, sx, sy);
	if (maxValue (image, x, y-1) > 0)
	  n += DoLine (x, y-1, tr, tg, sx, sy);
	}
  return n + x2-x1;
  }

long TMainWindow::Fill (int &x, int &y, long &tr, long &tg)
  {
  long sx=0, sy=0;
  tr=tg=0;
  long n =  DoLine (x, y, tr, tg, sx, sy);

  if (tr+tg > 0)
	{
	x = sx / (tr+tg);
	y = sy / (tr+tg);
	}

  return n;
  }

void TMainWindow::SetStatus(char *function, int percent)
  {
  if (LastPercent!=-1) StatText->SetText("100% Done");
  LastPercent=percent;
  if (function) FuncText->SetText(function);
	else FuncText->Clear();
  if (percent==-1) StatText->Clear();
	else EvTimer(0);
  }

void TMainWindow::EvTimer(UINT)
  {
  if ((LastPercent==-1)&&(numCells==-1)) return;
  if (LastPercent==99) LastPercent++;
  char c[40];

  if (LastPercent!=-1)
	{
	itoa (LastPercent, c, 10);
	strcat (c, "% Done");
	}
  else {
	itoa (numCells, c, 10);
	strcat (c, " Done");
	}

  StatText->SetText (c);
  }

void TMainWindow::EvSize (UINT sizeType, TSize &size)
  {
  fileNameLB->SetWindowPos (0, 0, 0, size.cx+2, size.cy+2-LByofs, SWP_NOZORDER | SWP_NOMOVE);
  TFrameWindow::EvSize (sizeType, size);
  }

void TMainWindow::MouseMovePosition(TPoint &point)
  {
  if (!image) return;
  if (prefs.displayRGBValues)
	{
	int r, g, b;
	char s[32];
	image->GetPixelTC(point.x, point.y, r, g, b);
	wsprintf(s, "red:%d green:%d blue:%d", r, g, b);
	StackText->SetText(s);
	}
  }


//  TDisplayWindow Methods

TDisplayWindow::TDisplayWindow (TWindow *AParent) :
	 TFrameWindow (AParent, "Displayed Image")
  {
  Attr.X = 100;  Attr.Y = 50;
  Attr.Style = WS_VISIBLE | WS_POPUP | WS_OVERLAPPEDWINDOW;
  displaying = 0;
  pal = 0;  dragDC = 0;  dib = 0;
  ClearLines();
  curline.numpoints = 0;
  strcpy (lineSaveName, "");
  MakePalette();
  handCount = 0;
  cellPoints = 0;
  circleSize = circleSizeSmall;
  }

TDisplayWindow::~TDisplayWindow ()
  {
  if (dragDC) delete dragDC;
  if (dib) delete dib;
  if (pal) delete pal;
  if (handCount) delete handCount;
  if (cellPoints) delete cellPoints;
  }

inline int TDisplayWindow::CanClose ()
  {  return 0;  }     //close only by closing parent window

void TDisplayWindow::MakePalette()
  {
  pal = 0;
  LPLOGPALETTE Palette = (LPLOGPALETTE) malloc (sizeof(LOGPALETTE) + sizeof(PALETTEENTRY)*224);
  if (!Palette) return;
  Palette->palVersion = 0x300;
  Palette->palNumEntries = 224;
  int r, g;
  for (r=0; r<14; r++)		// 14 reds
	for (g=0; g<16; g++)	// 16 greens (total 224)
	  Palette->palPalEntry[r*16+g].peRed = (r*256/14)+r,
	  Palette->palPalEntry[r*16+g].peGreen = (g<<4)+g,
	  Palette->palPalEntry[r*16+g].peBlue = 0,
	  Palette->palPalEntry[r*16+g].peFlags = 0;
  pal = new TPalette(Palette);
  free ((void *)Palette);
  }

void TDisplayWindow::Circle (int x, int y)
  {
  TClientDC dc(HWindow);
  dc.SelectObject(TPen(RGB(0xFF, 0xFF, 0xFF)));
  dc.Ellipse (x-2, y-2, x+2, y+2);
  dc.RestorePen();

//  if (dib)
//	{
//	TDibDC memDC(*dib);
//	memDC.SelectObject(TPen(RGB(0xFF, 0xFF, 0xFF)));
//	memDC.Ellipse (x-2, y-2, x+2, y+2);
//	memDC.RestorePen();
//	}
  }

void TDisplayWindow::DisplayImage (char *name, MemoryImage *image)
  {
  displaying++;
  TClientDC dc(HWindow);

  image->GetImageSize (w, h);
  int cx, cy;
  cx = w + 2 * GetSystemMetrics (SM_CXFRAME);
  cy = h + 2 * GetSystemMetrics(SM_CYFRAME) + GetSystemMetrics(SM_CYCAPTION);
  SetWindowPos (0, 0, 0, cx, cy-1, SWP_NOZORDER | SWP_NOMOVE);
  SetCaption (name);

  dc.SelectObject(*pal);
  dc.RealizePalette();

  if (dib) delete dib;
  dib = image->CreateDib(8, pal);
  if (!dib) return;

  dc.SetDIBitsToDevice(TRect(0,0,w,h), TPoint(0,0), *dib);

  displaying--;
  }

void TDisplayWindow::SetLineSaveName (char *name)
  {
  if (strcmp (lineSaveName, name)) ClearLines();
  strncpy (lineSaveName, name, fsPathName);
  }

void TDisplayWindow::ClearLines()
  {
  topline.numpoints = 0;
  boneline.numpoints = 0;
  s_iline.numpoints = 0;
  i_dline.numpoints = 0;
  leftline.numpoints = 0;
  rightline.numpoints = 0;
  }

void TDisplayWindow::Paint (TDC &dc, BOOL, TRect& rect)
  {
  if ((dib) && (!displaying))
	{
	if (pal) dc.SelectObject(*pal);
	dc.RealizePalette();
	if (rect.right > w) rect.right=w;
	if (rect.bottom > h) rect.bottom = h;
	dc.SetDIBitsToDevice(rect, TPoint(rect.left,h-rect.bottom), *dib);
	}
  int i;

  dc.SelectObject(TPen(RGB(0xFF, 0xFF, 0)));
  if (topline.numpoints)
	{
	dc.MoveTo(topline.points[0].x, topline.points[0].y);
	for (i=1; i<topline.numpoints; i++)
	  dc.LineTo(topline.points[i].x, topline.points[i].y);
	}
  if (s_iline.numpoints)
	{
	dc.MoveTo(s_iline.points[0].x, s_iline.points[0].y);
	for (i=1; i<s_iline.numpoints; i++)
	  dc.LineTo(s_iline.points[i].x, s_iline.points[i].y);
	}
  if (i_dline.numpoints)
	{
	dc.MoveTo(i_dline.points[0].x, i_dline.points[0].y);
	for (i=1; i<i_dline.numpoints; i++)
	  dc.LineTo(i_dline.points[i].x, i_dline.points[i].y);
	}
  if (boneline.numpoints)
	{
	dc.MoveTo (boneline.points[0].x, boneline.points[0].y);
	for (i=1; i<boneline.numpoints; i++)
	  dc.LineTo(boneline.points[i].x, boneline.points[i].y);
	}
  if (leftline.numpoints)
	{
	dc.MoveTo(leftline.points[0].x, leftline.points[0].y);
	for (i=1; i<leftline.numpoints; i++)
	  dc.LineTo(leftline.points[i].x, leftline.points[i].y);
	}
  if (rightline.numpoints)
	{
	dc.MoveTo(rightline.points[0].x, rightline.points[0].y);
	for (i=1; i<rightline.numpoints; i++)
	  dc.LineTo(rightline.points[i].x, rightline.points[i].y);
	}
  dc.RestorePen();

  if (cellPoints)
	{
	LOGBRUSH lb;  lb.lbStyle=BS_HOLLOW;  lb.lbColor=0;  lb.lbHatch=0;
	dc.SelectObject(TBrush(&lb));
	TCellPointsIterator j(*cellPoints);
	dc.SelectObject(TPen(RGB(0, 0xFF, 0)));  int isGreen=1;
	while (j)
	  {
	  TPoint p = j++;
	  if ((p.x>=0)&&(!isGreen))
		{
		dc.RestorePen();
		dc.SelectObject(TPen(RGB(0, 0xFF, 0)));  isGreen=1;
		}
	  if ((p.x<0)&&(isGreen))
		{
		dc.RestorePen();
		dc.SelectObject(TPen(RGB(0xFF, 0, 0)));  isGreen=0;
		}
      if (p.x<0) p.x = ~p.x;
	  dc.Ellipse (p.x-circleSize, p.y-circleSize, p.x+circleSize, p.y+circleSize);
	  }
	dc.RestoreBrush();
	}
  }

void TDisplayWindow::SaveLines ()
  {
  if (!strlen (lineSaveName)) return;
  ofstream *out = new ofstream (lineSaveName);
  int i;

  *out << "TOP LINE" << "\n";
  for (i=0; i<topline.numpoints; i++)
	*out << topline.points[i].x << "\t" << topline.points[i].y << "\n";
  *out << "SUP-INT LINE" << "\n";
  for (i=0; i<s_iline.numpoints; i++)
	*out << s_iline.points[i].x << "\t" << s_iline.points[i].y << "\n";
  *out << "INT-DEEP LINE" << "\n";
  for (i=0; i<i_dline.numpoints; i++)
	*out << i_dline.points[i].x << "\t" << i_dline.points[i].y << "\n";
  *out << "BONE LINE" << "\n";
  for (i=0; i<boneline.numpoints; i++)
	*out << boneline.points[i].x << "\t" << boneline.points[i].y << "\n";
  *out << "LEFT LINE" << "\n";
  for (i=0; i<leftline.numpoints; i++)
	*out << leftline.points[i].x << "\t" << leftline.points[i].y << "\n";
  *out << "RIGHT LINE" << "\n";
  for (i=0; i<rightline.numpoints; i++)
	*out << rightline.points[i].x << "\t" << rightline.points[i].y << "\n";

  delete out;
  }

void TDisplayWindow::UndoLastCell()
  {
  if (!cellPoints) return;
  int size = cellPoints->GetItemsInContainer();
  if (size)
	{
	if ((*cellPoints)[size-1].x < 0) redCount--;
	else greenCount--;
	cellPoints->RemoveEntry(size-1);
	handCount->SetCounts(redCount, greenCount);
	TMainWindow *mw = TYPESAFE_DOWNCAST(GetApplication()->GetMainWindow(), TMainWindow);
	mw->SetCount(cellPoints->GetItemsInContainer());
	Invalidate(FALSE);
	}
  }

void TDisplayWindow::EvRButtonDown(UINT, TPoint& point)
  {
  if (handCount)
	{
	TClientDC dc(*this);
	dc.SelectObject(TPen(RGB(0, 0xFF, 0)));
	LOGBRUSH lb;  lb.lbStyle=BS_HOLLOW;  lb.lbColor=0;  lb.lbHatch=0;
	dc.SelectObject(TBrush(&lb));
	dc.Ellipse (point.x-circleSize, point.y-circleSize, point.x+circleSize, point.y+circleSize);
	dc.RestorePen();
	dc.RestoreBrush();

	int i = cellPoints->Find(point);
	if (i!=INT_MAX) greenCount--;
	else {
	  i = cellPoints->Find(TPoint(~point.x,point.y));
	  if (i!=INT_MAX) redCount--;
	  }
	if (i==INT_MAX) cellPoints->Add(point);
	else cellPoints->SetData(i, point);

	greenCount++;

	handCount->SetCounts(redCount, greenCount);
	TMainWindow *mw = TYPESAFE_DOWNCAST(GetApplication()->GetMainWindow(), TMainWindow);
	mw->SetCount(cellPoints->GetItemsInContainer());
	}
  }

void TDisplayWindow::EvLButtonDown(UINT, TPoint& point)
  {
  if (handCount)
	{
	TClientDC dc(*this);
	dc.SelectObject(TPen(RGB(0xFF, 0, 0)));
	LOGBRUSH lb;  lb.lbStyle=BS_HOLLOW;  lb.lbColor=0;  lb.lbHatch=0;
	dc.SelectObject(TBrush(&lb));
	dc.Ellipse (point.x-circleSize, point.y-circleSize, point.x+circleSize, point.y+circleSize);
	dc.RestorePen();
	dc.RestoreBrush();

	int i = cellPoints->Find(point);
	if (i!=INT_MAX) greenCount--;
	else {
	  i = cellPoints->Find(TPoint(~point.x,point.y));
	  if (i!=INT_MAX) redCount--;
	  }
	point.x = ~point.x;
	if (i==INT_MAX) cellPoints->Add(point);
	else cellPoints->SetData(i, point);

	redCount++;

	handCount->SetCounts(redCount, greenCount);
	TMainWindow *mw = TYPESAFE_DOWNCAST(GetApplication()->GetMainWindow(), TMainWindow);
	mw->SetCount(cellPoints->GetItemsInContainer());
	return;
	}

  if (!dragDC)
	{
	SetCapture();
	dragDC = new TClientDC(HWindow);
	dragDC->MoveTo(point.x, point.y);
	curline.points[0].x = point.x;
	curline.points[0].y = point.y;
	curline.numpoints = 1;
	}
  }

void TDisplayWindow::EvMouseMove(UINT, TPoint& point)
  {
  if (dragDC)
	{
	dragDC->SelectObject(TPen(RGB(0, 0xFF, 0xFF)));
	dragDC->LineTo(point.x, point.y);
	int n = curline.numpoints;
	if ((n>1) && ((point.y-curline.points[n-1].y)*(point.x-curline.points[n-2].x)
	==(point.y-curline.points[n-2].y)*(point.x-curline.points[n-1].x)))
	  curline.numpoints--;
	curline.points[curline.numpoints].x = point.x;
	curline.points[curline.numpoints].y = point.y;
	curline.numpoints++;
	if (curline.numpoints >= maxpoints) curline.numpoints--;
	dragDC->RestorePen();
	}
  else {
	TMainWindow *mw = TYPESAFE_DOWNCAST(GetApplication()->GetMainWindow(), TMainWindow);
	mw->MouseMovePosition(point);
	}
  }

void TDisplayWindow::EvLButtonUp( UINT, TPoint& )
  {
  if (dragDC)
	{
	ReleaseCapture ();
	delete dragDC;  dragDC = NULL;

	TDialog *p;
	selectlinemsg m;
	p = new TDialog (this, dlg_SELECTLINE, NULL);
	p->SetTransferBuffer (&m);
	new TRadioButton (p, 101, NULL);  m.top   = 1;
	new TRadioButton (p, 102, NULL);  m.s_i   = 0;
	new TRadioButton (p, 103, NULL);  m.i_d   = 0;
	new TRadioButton (p, 104, NULL);  m.deep  = 0;
	new TRadioButton (p, 105, NULL);  m.left  = 0;
	new TRadioButton (p, 106, NULL);  m.right = 0;
	if (GetApplication()->ExecDialog(p) == IDOK)
	  {
	  if (m.top)    topline = curline;
	  if (m.s_i)    s_iline = curline;
	  if (m.i_d)    i_dline = curline;
	  if (m.deep)   boneline = curline;
	  if (m.left)   leftline = curline;
	  if (m.right)  rightline = curline;
	  SaveLines ();
	  }
	Invalidate();
	}
  }

void TDisplayWindow::StartHandCount()
  {
  redCount = greenCount = 0;
  handCount = new THandCountDialog(this);
  handCount->Create();
  cellPoints = new TCellPoints(256, 0, 128);
  }

void TDisplayWindow::DoneHandCount()
  {
  handCount = 0;
  TMainWindow *mw = TYPESAFE_DOWNCAST(GetApplication()->GetMainWindow(), TMainWindow);
  mw->DoneHandCounting(cellPoints);
  delete cellPoints;  cellPoints = 0;
  Invalidate();
  }


//  TPreferences methods

TPreferences::TPreferences()
  {
  char winDir[fsPathName];
  GetWindowsDirectory(winDir, fsPathName);
  GetProfileString("Analysis Tools", "INI-Path", winDir, iniFile, fsPathName);
  WriteProfileString("Analysis Tools", "INI-Path", iniFile);
  if ((strlen(iniFile) != 0) && (iniFile[strlen(iniFile)-1] != '\\') &&
	  (iniFile[strlen(iniFile)-1] != ':'))
	strcat(iniFile, "\\");
  strcat(iniFile, INIFileName);

  RestoreDefaults();
  }

void TPreferences::RestoreDefaults()
  {
  minCellSize = minCellSizeDefault;
  maxPixs = maxPixsDefault;
  defSaveEnhanced = saveEnhancedDefault;
  use24Bit = use24BitDefault;
  idealSat = idealSatDefault;
  DATPath[0] = 0;
  backgroundStDevs = backgroundStDevsDefault;
  displayStackUsage = stackUsageDefault;
  displayHistogram = histogramDefault;
  displayRGBValues = RGBValuesDefault;
  extraSmooth = extraSmoothDefault;
  localization = localizationDefault;

  thres = thresDefault;
  useOldEnhance = useOldEnhanceDefault;
  }

void TPreferences::LoadPreferences()
  {
  minCellSize = GetPrivateProfileInt(CartanalSection, "MinimumCellSize", minCellSizeDefault, iniFile);
  maxPixs = GetPrivateProfileInt(CartanalSection, "MaximumPixelsForColor", maxPixsDefault, iniFile);
  defSaveEnhanced = GetPrivateProfileInt(CartanalSection, "SaveEnhanced", saveEnhancedDefault, iniFile);
  use24Bit = GetPrivateProfileInt(CartanalSection, "Use24BitDIB", use24BitDefault, iniFile);
  idealSat = GetPrivateProfileInt(CartanalSection, "IdealSaturation", idealSatDefault, iniFile);
  backgroundStDevs = GetPrivateProfileInt(CartanalSection, "BackgroundStDevs", backgroundStDevsDefault, iniFile);
  GetPrivateProfileString(CartanalSection, "DAT-Path", ".", DATPath, fsPathName, iniFile);
  displayStackUsage = GetPrivateProfileInt(CartanalSection, "DisplayStackUsage", stackUsageDefault, iniFile);
  displayHistogram = GetPrivateProfileInt(CartanalSection, "DisplayHistogram", histogramDefault, iniFile);
  displayRGBValues = GetPrivateProfileInt(CartanalSection, "DisplayRGBValues", RGBValuesDefault, iniFile);
  extraSmooth = GetPrivateProfileInt(CartanalSection, "ExtraSmoothing", extraSmoothDefault, iniFile);
  localization = GetPrivateProfileInt(CartanalSection, "Localization", localizationDefault, iniFile);

  useOldEnhance = GetPrivateProfileInt(CartanalSection, "UseOldEnhance", 1, iniFile);
  }

void TPreferences::SavePreferences()
  {
  char s[16];
  wsprintf(s, "%d", minCellSize);
  WritePrivateProfileString("Cartilage Analysis", "MinimumCellSize", s, iniFile);
  wsprintf(s, "%d", maxPixs);
  WritePrivateProfileString("Cartilage Analysis", "MaximumPixelsForColor", s, iniFile);
  wsprintf(s, "%d", defSaveEnhanced);
  WritePrivateProfileString("Cartilage Analysis", "SaveEnhanced", s, iniFile);
  wsprintf(s, "%d", use24Bit);
  WritePrivateProfileString("Cartilage Analysis", "Use24BitDIB", s, iniFile);
  wsprintf(s, "%d", idealSat);
  WritePrivateProfileString("Cartilage Analysis", "IdealSaturation", s, iniFile);
  wsprintf(s, "%d", backgroundStDevs);
  WritePrivateProfileString("Cartilage Analysis", "BackgroundStDevs", s, iniFile);
  WritePrivateProfileString("Cartilage Analysis", "DAT-Path", DATPath, iniFile);
  wsprintf(s, "%d", displayStackUsage);
  WritePrivateProfileString("Cartilage Analysis", "DisplayStackUsage", s, iniFile);
  wsprintf(s, "%d", displayHistogram);
  WritePrivateProfileString("Cartilage Analysis", "DisplayHistogram", s, iniFile);
  wsprintf(s, "%d", displayRGBValues);
  WritePrivateProfileString("Cartilage Analysis", "DisplayRGBValues", s, iniFile);
  wsprintf(s, "%d", extraSmooth);
  WritePrivateProfileString("Cartilage Analysis", "ExtraSmoothing", s, iniFile);
  wsprintf(s, "%d", localization);
  WritePrivateProfileString("Cartilage Analysis", "Localization", s, iniFile);

  wsprintf(s, "%d", useOldEnhance);
  WritePrivateProfileString(CartanalSection, "UseOldEnhance", s, iniFile);
  }


//  TPreferencesDialog methods

TPreferencesDialog::TPreferencesDialog(TWindow* parent, TPreferences *Aprefs)
  : TDialog(parent, DLG_PREFERENCES)
  {
  prefs = Aprefs;
  MinCellSize = new TEdit(this, IDC_MINCELLSIZE, 10);
  MinCellSize->SetValidator(new TRangeValidator(1, 100));
  MaxPixs = new TEdit(this, IDC_MAXPIXS, 10);
  MaxPixs->SetValidator(new TRangeValidator(0, 10000));
  SaveEnh = new TCheckBox(this, IDC_SAVEENH);
  Use24Bit = new TCheckBox(this, IDC_24BIT);
  DATPath = new TEdit(this, IDC_DATPATH, fsPathName);
  BackStDevs = new TEdit(this, IDC_BACKSTDEVS, 10);
  BackStDevs->SetValidator(new TRangeValidator(1,50));
  ImageSat = new TEdit(this, IDC_IMAGESAT, 10);
  ImageSat->SetValidator(new TRangeValidator(10, 250));
  DisplayHistogram = new TCheckBox(this, IDC_HISTOGRAM);
  DisplayStackUsage = new TCheckBox(this, IDC_STACKUSAGE);
  DisplayRGBValues = new TCheckBox(this, IDC_RGBVALUES);
  ExtraSmooth = new TCheckBox(this, IDC_EXTRASMOOTH);
  Global = new TRadioButton(this, IDC_SIZELARGE);
  Medium = new TRadioButton(this, IDC_SIZEMED);
  Local = new TRadioButton(this, IDC_SIZESMALL);

  UseOldEnhance = new TCheckBox(this, IDC_OLDENHANCE);
  }

void TPreferencesDialog::SetupWindow()
  {
  TDialog::SetupWindow();
  char s[10];
  wsprintf(s, "%d", prefs->minCellSize);
  MinCellSize->SetText(s);
  wsprintf(s, "%d", prefs->maxPixs);
  MaxPixs->SetText(s);
  SaveEnh->SetCheck(prefs->defSaveEnhanced);
  Use24Bit->SetCheck(prefs->use24Bit);
  DATPath->SetText(prefs->DATPath);
  wsprintf(s, "%d", prefs->idealSat);
  ImageSat->SetText(s);
  wsprintf(s, "%d", prefs->backgroundStDevs);
  BackStDevs->SetText(s);
  DisplayHistogram->SetCheck(prefs->displayHistogram);
  DisplayStackUsage->SetCheck(prefs->displayStackUsage);
  DisplayRGBValues->SetCheck(prefs->displayRGBValues);
  ExtraSmooth->SetCheck(prefs->extraSmooth);
  switch (prefs->localization) {
	case 0: Global->SetCheck(1); break;
	case 1: Medium->SetCheck(1); break;
	case 2: Local->SetCheck(1);  break;
	}

  UseOldEnhance->SetCheck(prefs->useOldEnhance);
  }

void TPreferencesDialog::CmDefaults()
  {
  char s[16];
  wsprintf(s, "%d", minCellSizeDefault);
  MinCellSize->SetText(s);
  wsprintf(s, "%d", maxPixsDefault);
  MaxPixs->SetText(s);
  SaveEnh->SetCheck(saveEnhancedDefault);
  Use24Bit->SetCheck(use24BitDefault);
//  DATPath->SetText(".");
  wsprintf(s, "%d", idealSatDefault);
  ImageSat->SetText(s);
  wsprintf(s, "%d", backgroundStDevsDefault);
  BackStDevs->SetText(s);
  DisplayHistogram->SetCheck(histogramDefault);
  DisplayStackUsage->SetCheck(stackUsageDefault);
  DisplayRGBValues->SetCheck(RGBValuesDefault);
  ExtraSmooth->SetCheck(extraSmoothDefault);
  switch (localizationDefault) {
	case 0: Global->SetCheck(1); Medium->SetCheck(0); Local->SetCheck(0); break;
	case 1: Global->SetCheck(0); Medium->SetCheck(1); Local->SetCheck(0); break;
	case 2: Global->SetCheck(0); Medium->SetCheck(0); Local->SetCheck(1); break;
	}

  UseOldEnhance->SetCheck(useOldEnhanceDefault);
  }

void TPreferencesDialog::CmOk()
  {
  char s[16];
  MinCellSize->GetText(s, 16);  prefs->minCellSize = atoi(s);
  MaxPixs->GetText(s, 16);  prefs->maxPixs = atoi(s);
  prefs->defSaveEnhanced = SaveEnh->GetCheck();
  prefs->use24Bit = Use24Bit->GetCheck();
  ImageSat->GetText(s, 16);  prefs->idealSat = atoi(s);
  BackStDevs->GetText(s, 16);  prefs->backgroundStDevs = atoi(s);
  DATPath->GetText(prefs->DATPath, fsPathName);
  prefs->displayHistogram = DisplayHistogram->GetCheck();
  prefs->displayStackUsage = DisplayStackUsage->GetCheck();
  prefs->displayRGBValues = DisplayRGBValues->GetCheck();
  prefs->extraSmooth = ExtraSmooth->GetCheck();
  if (Local->GetCheck()) prefs->localization = 2;
  else if (Medium->GetCheck()) prefs->localization = 1;
  else prefs->localization = 0;

  prefs->useOldEnhance = UseOldEnhance->GetCheck();

  prefs->SavePreferences();
  TDialog::CmOk();
  }


//  THandCountDialog methods

THandCountDialog::THandCountDialog(TDisplayWindow* parent)
  : TDialog(parent, DLG_HANDCOUNT)
  {
  dw = parent;

  smallC = new TRadioButton(this, IDC_SIZESMALL);
  med = new TRadioButton(this, IDC_SIZEMED);
  large = new TRadioButton(this, IDC_SIZELARGE);
  red = new TStatic(this, IDC_RED, 10);
  green = new TStatic(this, IDC_GREEN, 10);
  }

void THandCountDialog::SetupWindow()
  {
  TDialog::SetupWindow();
  int size = dw->GetCircleSize();
  if (size <= circleSizeSmall) smallC->SetCheck(1);
  else if (size < circleSizeMedium) med->SetCheck(1);
  else large->SetCheck(1);
  red->SetText("0");
  green->SetText("0");
  }

void THandCountDialog::SetCounts(int redCount, int greenCount)
  {
  char s[10];
  wsprintf(s, "%d", redCount);
  red->SetText(s);
  wsprintf(s, "%d", greenCount);
  green->SetText(s);
  }

void THandCountDialog::BnCircleSize()
  {
  if (smallC->GetCheck()) dw->SetCircleSize(circleSizeSmall);
  else if (med->GetCheck()) dw->SetCircleSize(circleSizeMedium);
  else if (large->GetCheck()) dw->SetCircleSize(circleSizeLarge);
  }

void THandCountDialog::CmUndo()
  {
  dw->UndoLastCell();
  }

void THandCountDialog::CmOk()
  {
  dw->DoneHandCount();
  TDialog::CmOk();
  }


//  TApp methods

void TApp::InitMainWindow ()
  {
  SetMainWindow (new TMainWindow ());
  EnableCtl3d();
  }

void TApp::CmIndex()
  {
  GetMainWindow()->WinHelp(HLPFileName, HELP_INDEX, 0);
  }

void TApp::CmKeyboard()
  {
  GetMainWindow()->WinHelp(HLPFileName, HELP_INDEX, 0);
  }

void TApp::CmProcedures()
  {
  GetMainWindow()->WinHelp(HLPFileName, HELP_INDEX, 0);
  }

void TApp::CmUsingHelp()
  {
  GetMainWindow()->WinHelp("WINHELP.HLP", HELP_HELPONHELP, 0);
  }

void TApp::CmAbout()
  {
  TDialog (GetMainWindow(), dlg_ABOUT).Execute();
  }


//  THistogramWindow methods

DEFINE_RESPONSE_TABLE1(THistogramWindow, TFrameWindow)
  EV_WM_SIZE,
END_RESPONSE_TABLE;

THistogramWindow::THistogramWindow(TWindow *AParent, char *title, long *data, int Amean, int Astdev)
	: TFrameWindow(AParent, title)
  {
  int i;
  max=0;
  for (i=0; i<256; i++)
	{
	if (data[i]>max) max=data[i];
	hData[i]=data[i];
	}
  Attr.W = 270;
  Attr.H = 192;
  Attr.Style = WS_VISIBLE | WS_POPUP | WS_OVERLAPPEDWINDOW;
  mean = Amean;  stdev = Astdev;
  }

void THistogramWindow::Paint (TDC &PaintDC, BOOL, TRect &)
  {
  if (!max) return;
  TRect r;  GetClientRect(r);
  int i;
  for (i=0; i<256; i++)
	{
	PaintDC.MoveTo(i, r.bottom);
	PaintDC.LineTo(i, r.bottom-(hData[i]*(r.bottom-r.top)/max));
	}
  TPen mark(TColor(0,255,255));
  PaintDC.SelectObject(mark);
  if (mean != -1)
	{
	PaintDC.MoveTo(mean, r.bottom);
	PaintDC.LineTo(mean, r.bottom-5);
	}
  if (stdev != -1)
	{
	PaintDC.MoveTo(mean+stdev, r.bottom);
	PaintDC.LineTo(mean+stdev, r.bottom-3);
	if (mean-stdev > 0)
	  {
	  PaintDC.MoveTo(mean-stdev, r.bottom);
	  PaintDC.LineTo(mean-stdev, r.bottom-3);
	  }
	}
  PaintDC.RestorePen();
  }

void THistogramWindow::EvSize(uint sizeType, TSize& size)
  {
  TFrameWindow::EvSize(sizeType, size);
  Invalidate();
  }

// Main Program

#pragma argsused
int OwlMain (int argc, char* argv[])
  {
  return TApp("CartilageApp").Run();
  }

