#if !defined(__image_h)
#define __image_h

/* $Id: image.h,v 1.1.1.1 1995/08/16 15:33:56 cvs Exp $
 *
 * Copyright (c) 1996 University of Alberta
 *
 * May be copied or modified under the terms of the GNU General Public
 * License.  See COPYING file for more information.
 */

#include <owl\owlcore.h>
#include <owl\gdiobjec.h>
#include <fstream.h>


#define FLAGS_PUBLICMASK	0x00FF

// public flags
#define IMAGE_READ 			0x0010
#define IMAGE_WRITE 		0x0020
#define IMAGE_READWRITE		(IMAGE_READ|IMAGE_WRITE)
#define IMAGE_MULTIIMAGE 	0x0040

#define IMAGETYPE_RGB		0x0001
#define IMAGETYPE_GRAYSCALE	0x0002
#define IMAGETYPE_CLUT		0x0003

#define IMAGE_RANDOMREAD	0x0100
#define IMAGE_RANDOWWRITE	0x0200
#define IMAGE_RANDOWREADWRITE	(IMAGE_RANDOMREAD|IMAGE_RANDOMWRITE)

#define IMAGE_NOCOMPRESSION			0
#define IMAGE_DEFAULTCOMPRESSION	-1
#define BMP_COMPRESSION_BITFIELDS	3

#define COLORPLANE_RED		0x01
#define COLORPLANE_GREEN 	0x02
#define COLORPLANE_BLUE		0x04
#define COLORPLANE_GRAY		0x08
typedef enum tColorPlane {gray=COLORPLANE_GRAY, red=COLORPLANE_RED,
						  green=COLORPLANE_GREEN, blue=COLORPLANE_BLUE };


//  Structures

struct TGAHeader;
struct TGAExtension;
struct TGADeveloper;
struct PCXHeader;
struct PCXFooter;
struct BITMAPARRAYHEADER;

struct RGBHistogram {
  long red[256], green[256], blue[256];
  };


//  Global Methods

class ImageFile;
ImageFile* DetermineImageTypeAndOpen(char *fname, int flags=IMAGE_READ);


//  Classes

class Image {
  protected:
	int imageType;
	TPalette *palette;
	int flags;
	void (far *StatusFunc) (int);

  public:
	Image (int aFlags=0) { palette=0; flags=aFlags; StatusFunc=0; }
	virtual ~Image () { if (palette) delete palette; }

	virtual void GetPixelRGB(int x, int y, int& r, int& g, int& b) { }
	virtual void PutPixelRGB(int x, int y, int r, int g, int b) { }
	virtual void GetPixelGray (int x, int y, int& g) { }
	virtual void PutPixelGray (int x, int y, int g) { }
	virtual void GetPixelCLUT (int x, int y, int& i) { }
	virtual void PutPixelCLUT (int x, int y, int i) { }
	virtual void ReadLineRGB (int y, unsigned char *buf) { }
	virtual void WriteLineRGB (int y, unsigned char *buf) { }

	virtual void GetImageSize (int &width, int &height) { }
	virtual int SetImageSize (int width, int height)=0;
	virtual void GetImageType (int &type) { }
	virtual void SetImageType (int type) { }

	virtual void SetPalette(const TPalette &Apalette);
	virtual const TPalette* GetPalette()
	  { return palette; }

	virtual BOOL CopyFrom(Image *AImage);
	virtual int GetOrientation()=0;

// obsolete functions
	enum {TRUECOLOR, GRAYSCALE, CLUT=4};
	inline void ReadLineTC (int y, unsigned char *buf) { ReadLineRGB(y,buf); }
	inline void WriteLineTC (int y, unsigned char *buf) { WriteLineRGB(y,buf); }
	inline BOOL LoadImage (Image *AImage) { return CopyFrom(AImage); }
	inline void GetPixelTC (int x, int y, int& r, int& g, int& b) { GetPixelRGB(x,y,r,g,b); }
	inline void PutPixelTC (int x, int y, int r, int g, int b) { PutPixelRGB(x,y,r,g,b); }

  protected:
	virtual void GetPixel (int x, int y, long& c) { }
	virtual void PutPixel (int x, int y, long c) { }
  };


class ImageFile : public Image {
  protected:
	LPVOID image;

	virtual BOOL AdjustFileSize(int newsize);

  private:
	HANDLE file;
	HANDLE fmapping;
	int filesize;

  public:
	ImageFile(char *fname, int flags);
	virtual ~ImageFile ();

	virtual BOOL SelectImage(int index=0)=0;
	virtual BOOL CreateNewImageAndSelect()=0;
	inline int GetCurrentFileSize() {return filesize;}

	inline int bad() { return image ? 0 : 1; }

	virtual int GetLineSize() = 0;
	virtual void GetPixelSize (int& bits, int& planes) = 0;
	virtual void SetPixelSize (int bits, int planes) = 0;

	virtual void ReadColorMap() {}
	virtual void WriteColorMap() {}

//  obsolete stuff
	virtual void CreateFooter() { }
	virtual void WriteHeader() { }
	enum {READ=IMAGE_READ, WRITE=IMAGE_WRITE};
  };


class TGAFile : public ImageFile {
  protected:
	TGAHeader *header;
	TGAExtension *extensionArea;
	TGADeveloper *developerArea;
	uint8 *bits;
	int newTGA;
	int readLinePacket;

  public:
	TGAFile(char *fname, int access);
	~TGAFile ();

	BOOL SelectImage(int index=0);

	int GetLineSize();
	void GetImageSize (int& width, int& height);
	void GetPixelSize (int& bits, int& planes);
	int SetImageSize (int width, int height);
	void SetPixelSize (int bits, int planes);
	int GetOrientation();
	BOOL CreateNewImageAndSelect() {return 0;}

	virtual void CreateFooter();
	virtual void ReadLineRGB (int y, unsigned char *buf);
	virtual void WriteLineRGB (int y, unsigned char *buf);
	virtual void WriteHeader();

	static int IsExtensionThisType(char *extension)
	  { if (stricmp(extension, "TGA")==0) return 1;
		return 0; }
	static int IsFileThisType(fstream *file);
  };


class PCXFile : public ImageFile {
  protected:
	PCXHeader *header;
	PCXFooter *footer;
  public:
	PCXFile(char *fname, int access);
	~PCXFile ();

	BOOL SelectImage(int index=0);

	int GetLineSize();
	void GetImageSize (int& width, int& height);
	void GetPixelSize (int& bits, int& planes);
	int SetImageSize (int width, int height);
	void SetPixelSize (int bits, int planes);
	int GetOrientation() { return 0; }
	BOOL CreateNewImageAndSelect() {return 0;}

	virtual void GetImageType (int &type);
	virtual void SetImageType (int type);

	virtual void CreateFooter();
	virtual void Read (unsigned char *buf, int lines);
	virtual void Write (unsigned char *buf, int lines);
	virtual void WriteHeader();
	virtual void ReadLineRGB (int y, unsigned char *buf);
	virtual void WriteLineRGB (int y, unsigned char *buf);

	static int IsExtensionThisType(char *extension)
	  { if (stricmp(extension, "PCX")==0) return 1;
		return 0; }
	static int IsFileThisType(fstream *file);
  };

class BMPFile : public ImageFile {
  protected:
	BITMAPARRAYHEADER *BMAH;
	BITMAPFILEHEADER *BMFH;
	BITMAPINFOHEADER *BMIH;
	uint8 *bits;
	int numImages;
	int redMask, greenMask, blueMask;

	virtual BOOL AdjustFileSize(int newsize);

  private:
	void CountImages();
	void CompleteHeader();

  public:
	BMPFile (char *fname, int access);
	~BMPFile () {}

	BOOL SelectImage(int index);
	BOOL CreateNewImageAndSelect();

	void ReadLineRGB (int y, unsigned char *buf);
	void WriteLineRGB (int y, unsigned char *buf);
	void WriteLineCLUT (int y, unsigned char *buf);

	int GetLineSize();
	void GetImageSize (int& width, int& height);
	void GetPixelSize (int& bits, int& planes);
	int GetOrientation();

	int SetImageSize(int width, int height);
	void SetPixelSize(int bits, int planes);
	int SetCompression(int compressionType, int redBits=0, int greenBits=0, int blueBits=0);

	void ReadColorMap();
	void WriteColorMap();

	int SetSpecialFileType(long type);
	long GetSpecialFileType();

	static int IsExtensionThisType(char *extension)
	  { if ((stricmp(extension, "BMP")==0)||(stricmp(extension,"DIB")==0)) return 1;
		return 0; }
	static int IsFileThisType(fstream *file);
  };


class MemoryImage : public Image{
  protected:
	unsigned char far* image;
	int w, h, bitsPerPixel, redBits, greenBits, blueBits;
	RGBHistogram *histogram;

// Windows Data
	HGLOBAL imageHandle;

  public:
	MemoryImage (int rBits, int gBits, int bBits);
	MemoryImage (int grayBits);
	virtual ~MemoryImage ();

	virtual void GetPixelRGB(int x, int y, int& r, int& g, int& b);
	virtual void PutPixelRGB(int x, int y, int r, int g, int b);
	virtual void GetPixelGray (int x, int y, int& g);
	virtual void PutPixelGray (int x, int y, int g);
	virtual void GetPixelCLUT (int x, int y, int& i);
	virtual void PutPixelCLUT (int x, int y, int i);
	virtual void ReadLineRGB (int y, unsigned char *buf);
	virtual void WriteLineRGB (int y, unsigned char *buf);

	virtual int CreateImage (int width, int height);
	virtual void ClearImage (tColorPlane tc, int color);
	virtual void AcquireImage ();
	virtual void DisplayImage (int xOfs, int yOfs); // DOS only
	virtual TDib* CreateDib(int bitsPerPixel = 24, TPalette *pal = 0); // Windows Only
	virtual int GetOrientation() { return 0; }

	virtual void SetStatusFunction (void (far *statfunc)(int))
	  { StatusFunc = statfunc; }
	virtual void GetImageSize (int &width, int &height)
	  { width = w;  height = h; }
	int SetImageSize(int width, int height)
	  { if (!image) return CreateImage(width, height); return 0; }
	virtual void GetPixelSize (int &numBits)
	  { numBits = bitsPerPixel; }
	virtual void GetPixelSize (int &rBits, int &gBits, int &bBits)
	  { rBits = redBits;  gBits = greenBits;  bBits = blueBits; }
	virtual void GetImageType (int &type)
	  { type = imageType; }
	virtual void SetImageType (int type);

	virtual void Smooth (tColorPlane cp, int matrixSize,
			 int x1, int y1, int x2, int y2);
	virtual void Smooth (tColorPlane cp, int matrixSize=3);
	virtual void MatrixFilter (tColorPlane cp, int matrix[], int matrixSize,
			 int x1, int y1, int x2, int y2);
	virtual void MatrixFilter (tColorPlane cp, int matrix[], int matrixSize);
	virtual void AddImage (class MemoryImage *Aimage);
	virtual void AverageImage (class MemoryImage *Aimage);
	virtual void SubtractImage (class MemoryImage *Aimage);
	virtual void SubtractImageReverse (class MemoryImage *Aimage);
	virtual void DivideContrast (tColorPlane cp, int factor);
	virtual void MultiplyContrast (tColorPlane cp, int factor);
	virtual void ShiftIntensity (tColorPlane cp, int cnst);
	virtual void EnhanceContrast (tColorPlane cp, int low, int high);
	virtual void ThresholdImage (tColorPlane cp, int thres, int low, int high);
	virtual void ReverseImage (tColorPlane cp);
	virtual void DrawContours (tColorPlane cp, int num, int color);
	virtual void DrawContours (tColorPlane cp, int num, int color,
			 class MemoryImage *AImage);

	virtual void CalculateHistogram ();
	virtual void SmoothHistogram ();
	virtual long GetHistogramValue (tColorPlane cp, int intensity);
	virtual int GetMinimumIntensity (tColorPlane cp);
	virtual int GetMaximumIntensity (tColorPlane cp);
	virtual int GetBackgroundIntensity (tColorPlane cp);

  protected:
	virtual void GetPixel (int x, int y, long& c);
	virtual void PutPixel (int x, int y, long c);
	virtual void PutPixel (tColorPlane cp, int x, int y, long c);
  };

#endif

