/* $Id: imagewin.cpp,v 1.1.1.1 1995/07/31 13:53:08 cvs Exp $
 *
 * Copyright (c) 1996 University of Alberta
 *
 * May be copied or modified under the terms of the GNU General Public
 * License.  See COPYING file for more information.
 */

#include <owl\owlcore.h>
#include <alloc.h>
#pragma hdrstop

#include "image.h"

MemoryImage::MemoryImage (int rBits, int gBits, int bBits) : Image ()
  {
  imageType = TRUECOLOR;
  redBits = rBits;
  greenBits = gBits;
  blueBits = bBits;
  bitsPerPixel = rBits+gBits+bBits;
  if (bitsPerPixel <= 24)
	while (!bitsPerPixel == (1, 2, 4, 8, 16, 24)) bitsPerPixel++;
  image = 0;  imageHandle = 0;
  histogram = 0;
  }

MemoryImage::MemoryImage (int grayBits) : Image ()
  {
  imageType = GRAYSCALE;
  bitsPerPixel = grayBits;
  if (bitsPerPixel <= 24)
	while (!bitsPerPixel==(1, 2, 4, 8, 16, 24)) bitsPerPixel++;
  redBits = greenBits = blueBits = 0;
  image = 0;  imageHandle = 0;
  histogram = 0;
  }

MemoryImage::~MemoryImage ()
  {
  if (imageHandle)
	{
	GlobalUnlock (imageHandle);
	GlobalFree (imageHandle);
	}
  if (histogram) delete histogram;
  }

int MemoryImage::CreateImage (int width, int height)
  {
  if (imageHandle)
	{
	GlobalUnlock (imageHandle);
	GlobalFree (imageHandle);
	image = 0;
	}
  h = height;
  if (bitsPerPixel == 1) w = (width+7)&0xFFF8;
	else w = (width+3)&0xFFFC;

  DWORD size = (w * bitsPerPixel);
  size = (((h * size) + 7) >> 3) + 2;
  imageHandle = GlobalAlloc (GMEM_MOVEABLE, size);
  if (imageHandle) image = (unsigned char far*)GlobalLock(imageHandle);
  return (image ? 1 : 0);
  }

TDib* MemoryImage::CreateDib(int colorDepth, TPalette* pal)
  {
  if (!colorDepth==(8,24)) return 0;
  if ((colorDepth==8) && (!pal))
	{ if (palette) pal=palette; else return 0; }

  int wd = (w+3)&0xFFFC;
  TDib *dib = new TDib(wd, h, 256, DIB_PAL_COLORS);
  if (!dib) return 0;
  unsigned char* colortable = new unsigned char[0x1000];
  if (!colortable)
	{ delete dib; return 0; }
  unsigned char* bits = (unsigned char*)dib->GetBits();
  if (!bits)
	{ delete dib; delete colortable; return 0; }
  if (StatusFunc) StatusFunc (0);

  int i,r,g,b;
  for (i=0; i<0x1000; i++)
	{
	r=(i&0x00F)<<4;  g=(i&0x0F0);  b=(i&0xF00)>>4;
	colortable[i] = pal->GetNearestPaletteIndex(TColor(r,g,b));
	}

  PALETTEENTRY pe;
  for (i=0; i<pal->GetNumEntries(); i++)
	{
	pal->GetPaletteEntry(i, pe);
	dib->SetColor(i, TColor(pe.peRed,pe.peGreen,pe.peBlue));
	}

  int x, y, pos=0;
  switch (imageType) {
	case (CLUT):
	case (GRAYSCALE):
	case (TRUECOLOR):
	  for (y=0; y<h; y++)
		{
		if (StatusFunc) StatusFunc ((long)100*y/h);
		for (x=0; x<w; x++)
		  {
		  GetPixelTC (x, h-y-1, r, g, b);
		  i=((r&0xF0)>>4) + (g&0xF0) + ((b&0xF0)<<4);
		  bits[pos++] = colortable[i];
		  }
		pos += (wd-w);
		}
	  break;

	}
  delete colortable;
  return dib;
  }

/*
HBITMAP MemoryImage::CreateBitmap (HDC dc)
  {
  int wd = (w+3)&0xFFFC;
  BITMAPINFO *bmi = (BITMAPINFO*) malloc (sizeof(BITMAPINFOHEADER)+256*sizeof(RGBQUAD));
  bmi->bmiHeader.biSize = sizeof (BITMAPINFOHEADER);
  bmi->bmiHeader.biWidth = wd;
  bmi->bmiHeader.biPlanes = 1;
  bmi->bmiHeader.biCompression = BI_RGB;
  bmi->bmiHeader.biXPelsPerMeter = 300;
  bmi->bmiHeader.biYPelsPerMeter = 300;
  HBITMAP bm = 0;

  int x, y, r, g, b;
  switch (imageType) {
	case (TRUECOLOR) :
	  bmi->bmiHeader.biHeight = h;
	  bmi->bmiHeader.biBitCount = 24;
	  bmi->bmiHeader.biSizeImage = ((long)h)*wd*3;
	  bmi->bmiHeader.biClrUsed = 0;
	  bmi->bmiHeader.biClrImportant = 0;
	  bm = CreateCompatibleBitmap (dc, wd, h);
	  if (!bm) break;
	  if ((redBits==8) && (greenBits==8) && (blueBits==8))
	{
	if (StatusFunc) StatusFunc (0);
	HCURSOR oldCur = SetCursor (LoadCursor (NULL, IDC_WAIT));
	SetDIBits (dc, bm, 0, h, (char far *)image, bmi, DIB_RGB_COLORS);
	SetCursor (oldCur);
	if (StatusFunc) StatusFunc (100);
	} else
	{
	int i=0, sy=0;
	HANDLE hMEM = GlobalAlloc (GMEM_MOVEABLE, ((long)30)*wd);
	if (!hMEM) return 0;
	unsigned char far* bits = (unsigned char far*) GlobalLock (hMEM);
	for (y=0; y<h; y++)
	  {
	  if (StatusFunc) StatusFunc ((long)100*y/h);
	  for (x=0; x<w; x++)
		{
		GetPixelTC (x, h-y-1, r, g, b);
		bits[i++] = b;
		bits[i++] = g;
		bits[i++] = r;
		};
	  i += 3*(wd-w);
	  if ((y%10)==9)
		{
		SetDIBits (dc, bm, sy, 10, (char far *)bits, bmi, DIB_RGB_COLORS);
		sy += 10;  i = 0;
		};
	  };
	if (h%10)
	  SetDIBits (dc, bm, sy, h-sy, (char far *)bits, bmi, DIB_RGB_COLORS);
	GlobalUnlock (hMEM);
	GlobalFree (hMEM);
	}
	  break;
	case (GRAYSCALE) :
	  bmi->bmiHeader.biHeight = h;
	  bmi->bmiHeader.biBitCount = 8;
	  bmi->bmiHeader.biSizeImage = ((long)wd)*h;
	  bmi->bmiHeader.biClrUsed = 0;
	  bmi->bmiHeader.biClrImportant = 0;
	  for (y=0; y<256; y++)
	bmi->bmiColors[y].rgbBlue  = y,
	bmi->bmiColors[y].rgbGreen = y,
	bmi->bmiColors[y].rgbRed   = y,
	bmi->bmiColors[y].rgbReserved = 0;
	  bm = CreateCompatibleBitmap (dc, wd, h);
	  if (!bm) break;

	  if (bitsPerPixel == 8)
	{
	if (StatusFunc) StatusFunc (0);
	HCURSOR oldCur = SetCursor (LoadCursor (NULL, IDC_WAIT));
	SetDIBits (dc, bm, 0, h, (char far *)image, bmi, DIB_RGB_COLORS);
	SetCursor (oldCur);
	if (StatusFunc) StatusFunc (100);
	} else
	{
	HANDLE hMEM = GlobalAlloc (GMEM_MOVEABLE, ((long)wd)*h);
	if (!hMEM) return 0;
	unsigned char far* bits = (unsigned char far*) GlobalLock (hMEM);
	for (y=0; y<h; y++)
	  {
	  if (StatusFunc) StatusFunc (((long)100)*y/h);
	  for (x=0; x<w; x++)
		{
		GetPixelGray (x, h-y-1, g);
		bits[(((long)y)*wd)+x] = g;
		}
	  }
	SetDIBits (dc, bm, 0, h, (char far *)bits, bmi, DIB_RGB_COLORS);
	GlobalUnlock (hMEM);
	GlobalFree (hMEM);
	}
	}
  return bm;
  }
*/

void MemoryImage::DisplayImage (int, int) { }

