/* $Id: tgacomp.cpp,v 1.1.1.1 1995/08/09 11:00:40 cvs Exp $
 *
 * Copyright (c) 1995 University of Alberta
 *
 * May be copied or modified under the terms of the GNU General Public
 * License.  See COPYING file for more information.
 */

#include <stdio.h>
#include <image.h>
#pragma hdrstop


void DisplayHelp()
  {
  printf("Usage:\n");
  printf("  cartcomp [-ex] infile outfile\n\n");
  exit(1);
  }

void ReportError(char *error)
  {
  printf("\n***ERROR*** %s\n\n", error);
  exit(2);
  }

int CopyImageNoBlue(Image *out, Image *in)
  {
  int w, h, y, i;

  const TPalette *pal = in->GetPalette();
  if ((!out->GetPalette()) && (pal)) out->SetPalette(*pal);

  in->GetImageSize (w, h);
  if (!out->SetImageSize (w, h)) return 0;
  unsigned char * line = (unsigned char *) malloc (w*3);
  if (!line) return 0;

  if (in->GetOrientation()&1)
	for (y=h-1; y>=0; y--)
	  {
//	  if (StatusFunc) StatusFunc ((long)100*(h-y-1)/h);
	  in->ReadLineRGB (y, line);
	  for (i=0; i<w; i++) line[i*3+2]=0;
	  out->WriteLineRGB (y, line);
	  }
  else for (y=0; y<h; y++)
	{
//	if (StatusFunc) StatusFunc ((long)100*y/h);
	in->ReadLineRGB (y, line);
    for (i=0; i<w; i++) line[i*3+2]=0;
	out->WriteLineRGB (y, line);
	}
  free (line);
  return 1;
  }

main (int argc, char*argv[])
  {
  printf("Cartilage Image Compression Program v1.0\n");
  printf("by Chris Studholme  (c)1995\n\n");
  if (argc<3) DisplayHelp();
  char *inFile=0, *outFile=0;
  int xtraCompress=0;

  argc--;
  while (argc)
	{
	if ((argv[argc][0]!='-')&&(argv[argc][0]!='/'))
	  {
	  if (!outFile) outFile = argv[argc];
	  else if (!inFile) inFile = argv[argc];
	  else DisplayHelp();
	  }
	else {
	  if (stricmp(&(argv[argc][1]),"ex")==0) xtraCompress=1;
	  else DisplayHelp();
	  }
	argc--;
	}
  if (!inFile) DisplayHelp();

  printf("Opening source file %s...", inFile);
  ImageFile *in = DetermineImageTypeAndOpen(inFile);
  if ((!in)||(in->bad())) ReportError("Unable to open source file");

  printf("\nOpening output file %s...", outFile);
  BMPFile *out = new BMPFile(outFile, IMAGE_WRITE);
  if ((!out)||(out->bad())) ReportError("Unable to open output file");

  if (!xtraCompress)
	out->SetPixelSize (24, 1);
  else {
	out->SetPixelSize (16, 1);
	out->SetCompression(BMP_COMPRESSION_BITFIELDS, 8, 8, 0);
	}

  printf("\nCoping image...");
//  out->CopyFrom(in);
  CopyImageNoBlue(out, in);
  delete out;
  delete in;
  printf("\nDone.\n\n");

  return 0;
  }
